# -*- coding: utf-8 -*-

# EnPT, EnMAP Processing Tool - A Python package for pre-processing of EnMAP Level-1B data
#
# Copyright (C) 2019  Karl Segl (GFZ Potsdam, segl@gfz-potsdam.de), Daniel Scheffler
# (GFZ Potsdam, danschef@gfz-potsdam.de), Niklas Bohn (GFZ Potsdam, nbohn@gfz-potsdam.de),
# Stéphane Guillaso (GFZ Potsdam, stephane.guillaso@gfz-potsdam.de)
#
# This software was developed within the context of the EnMAP project supported
# by the DLR Space Administration with funds of the German Federal Ministry of
# Economic Affairs and Energy (on the basis of a decision by the German Bundestag:
# 50 EE 1529) and contributions from DLR, GFZ and OHB System AG.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version. Please note the following exception: `EnPT` depends on tqdm, which
# is distributed under the Mozilla Public Licence (MPL) v2.0 except for the files
# "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""EnPT 'radiometric transform' module.

Contained Transformations:
    - TOA radiance to TOA reflectance
"""

import math
import numpy as np

from ...model.images import EnMAPL1Product_SensorGeo, EnMAP_Detector_SensorGeo  # noqa F401  # flake8 issue
from ...options.config import EnPTConfig

__author__ = 'Daniel Scheffler'


class Radiometric_Transformer(object):
    """Class for performing all kinds of radiometric transformations of EnMAP images."""

    def __init__(self, config: EnPTConfig = None):
        """Create an instance of Radiometric_Transformer.

        :param config: EnPT configuration object
        """
        self.cfg = config
        self.solarIrr = config.path_solar_irr  # path of model for solar irradiance
        self.earthSunDist = config.path_earthSunDist  # path of model for earth sun distance

    def transform_TOARad2TOARef(self, enmap_ImageL1: EnMAPL1Product_SensorGeo):
        """Transform top-of-atmosphere radiance to top-of-atmosphere reflectance (16-bit signed-integer).

        NOTE: The following formula is used:
                toaRef = (scale_factor * math.pi * toaRad * earthSunDist**2) /
                         (solIrr * math.cos(zenithAngleDeg))

        :param enmap_ImageL1:   instance of the class 'EnMAPL1Product_ImGeo'
        :return:
        """
        for detectorName in enmap_ImageL1.detector_attrNames:
            detector: EnMAP_Detector_SensorGeo = getattr(enmap_ImageL1, detectorName)

            enmap_ImageL1.logger.info('Converting TOA radiance to TOA reflectance for %s detector...'
                                      % detector.detector_name)

            # compute TOA reflectance
            constant = \
                self.cfg.scale_factor_toa_ref * math.pi * enmap_ImageL1.meta.earthSunDist ** 2 / \
                (math.cos(math.radians(enmap_ImageL1.meta.geom_sun_zenith)))
            solIrr = detector.detector_meta.solar_irrad.reshape(1, 1, detector.data.bands)
            toaRef = (constant * detector.data[:] / solIrr).astype(np.int16)

            # update EnMAP image
            detector.data = toaRef
            detector.detector_meta.unit = '0-%d' % self.cfg.scale_factor_toa_ref
            detector.detector_meta.unitcode = 'TOARef'

        return enmap_ImageL1



# -*- coding: utf-8 -*-

# EnPT, EnMAP Processing Tool - A Python package for pre-processing of EnMAP Level-1B data
#
# Copyright (C) 2019  Karl Segl (GFZ Potsdam, segl@gfz-potsdam.de), Daniel Scheffler
# (GFZ Potsdam, danschef@gfz-potsdam.de), Niklas Bohn (GFZ Potsdam, nbohn@gfz-potsdam.de),
# Stéphane Guillaso (GFZ Potsdam, stephane.guillaso@gfz-potsdam.de)
#
# This software was developed within the context of the EnMAP project supported
# by the DLR Space Administration with funds of the German Federal Ministry of
# Economic Affairs and Energy (on the basis of a decision by the German Bundestag:
# 50 EE 1529) and contributions from DLR, GFZ and OHB System AG.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version. Please note the following exception: `EnPT` depends on tqdm, which
# is distributed under the Mozilla Public Licence (MPL) v2.0 except for the files
# "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""EnPT metadata objects for EnMAP data in sensor geometry."""

from datetime import datetime
from lxml import etree as ElementTree
import logging
import os
import fnmatch
from typing import Union, List, Tuple, Optional  # noqa: F401
from collections import OrderedDict
import numpy as np
from py_tools_ds.geo.vector.topology import Polygon, get_footprint_polygon  # noqa: F401  # flake8 issue
from geoarray import GeoArray

from ...options.config import EnPTConfig
from ..srf import SRF
from ...processors.spatial_transform import RPC_3D_Geolayer_Generator

__author__ = ['Daniel Scheffler', 'Stéphane Guillaso', 'André Hollstein']


class EnMAP_Metadata_L1B_Detector_SensorGeo(object):
    """Class for all EnMAP metadata associated with a single EnMAP detector in sensor geometry.

    NOTE:
        - All metadata that have VNIR and SWIR detector in sensor geometry in common should be included here.

    """

    def __init__(self, detector_name: str, config: EnPTConfig, logger: logging.Logger = None):
        """Get a metadata object containing the metadata of a single EnMAP detector in sensor geometry.

        :param detector_name:   Name of the detector (VNIR or SWIR)
        :param config:          EnPT configuration object
        :param logger:          instance of logging.logger or subclassed
        """
        from . import L1B_product_props, L1B_product_props_DLR
        self.cfg = config
        self.detector_name: str = detector_name
        if not self.cfg.is_dummy_dataformat:
            self.detector_label = L1B_product_props_DLR['xml_detector_label'][detector_name]
        else:
            self.detector_label = L1B_product_props['xml_detector_label'][detector_name]
        self.logger = logger or logging.getLogger()

        # These lines are used to load path information
        self.filename_data: Optional[str] = ''  # detector data filename
        self.scene_basename: Optional[str] = ''  # basename of the EnMAP image
        self.filename_deadpixelmap: Optional[str] = ''  # filename of the dead pixel file
        self.filename_quicklook: Optional[str] = ''  # filename of the quicklook file
        # FIXME masks of BOTH detectors
        self.filename_mask_landwater: Optional[str] = ''  # filename of the land/water mask file
        self.filename_mask_snow: Optional[str] = ''  # filename of the snow mask file
        self.filename_mask_cloudshadow: Optional[str] = ''  # filename of the cloud shadow mask file
        self.filename_mask_clouds: Optional[str] = ''  # filename of the cloud mask file
        self.filename_mask_haze: Optional[str] = ''  # filename of the haze mask file
        self.filename_mask_cirrus: Optional[str] = ''  # filename of the cirrus mask file

        self.wvl_center: Optional[np.ndarray] = None  # Center wavelengths for each EnMAP band
        self.fwhm: Optional[np.ndarray] = None  # Full width half maximmum for each EnMAP band
        self.srf: Optional[SRF] = None  # SRF object holding the spectral response functions for each EnMAP band
        self.solar_irrad: Optional[np.ndarray] = None  # solar irradiance in [W/m2/nm] for each band
        self.nwvl: Optional[int] = None  # Number of wave bands
        self.nrows: Optional[int] = None  # number of rows
        self.ncols: Optional[int] = None  # number of columns
        self.smile_coef: Optional[np.ndarray] = None  # smile coefficients needed for smile computation
        self.nsmile_coef: Optional[int] = None  # number of smile coefficients
        self.smile: Optional[np.ndarray] = None  # smile for each EnMAP image column
        self.gains: Optional[np.ndarray] = None  # band-wise gains for computing radiance from DNs
        self.offsets: Optional[np.ndarray] = None   # band-wise offsets for computing radiance from DNs
        self.l_min: Optional[np.ndarray] = None  # band-wise l-min for computing radiance from DNs
        self.l_max: Optional[np.ndarray] = None  # band-wise l-max for computing radiance from DNs
        self.lat_UL_UR_LL_LR: Optional[List[float, float, float, float]] = None  # latitude coords for UL, UR, LL, LR
        self.lon_UL_UR_LL_LR: Optional[List[float, float, float, float]] = None  # longitude coords for UL, UR, LL, LR
        self.epsg_ortho: Optional[int] = None  # EPSG code of the orthorectified image
        self.rpc_coeffs: OrderedDict = OrderedDict()  # RPC coefficients for geolayer computation
        self.ll_mapPoly: Optional[Polygon] = None  # footprint polygon in longitude/latitude map coordinates
        self.lats: Optional[np.ndarray] = None  # 2D array of latitude coordinates according to given lon/lat sampling
        self.lons: Optional[np.ndarray] = None  # 2D array of longitude coordinates according to given lon/lat sampling
        self.unit: str = ''  # radiometric unit of pixel values
        self.unitcode: str = ''  # code of radiometric unit
        self.preview_wvls: Optional[List[float]] = None  # wavelengths to be used for quicklook images
        self.preview_bands: Optional[List[int]] = None  # band indices to be used for quicklook images
        self.snr: Optional[np.ndarray] = None   # Signal to noise ratio as computed from radiance data

    def read_metadata(self, path_xml):
        """
        Read the metadata of a specific EnMAP detector in sensor geometry

        :param path_xml: file path of the metadata file
        :return: None
        """
        xml = ElementTree.parse(path_xml).getroot()

        if not self.cfg.is_dummy_dataformat:
            lbl = self.detector_label
            self.logger.info("Reading metadata for %s detector..." % self.detector_name)

            # read data filenames
            all_filenames = [ele.text for ele in xml.findall("product/productFileInformation/file/name")]

            def get_filename(matching_exp: str):
                matches = fnmatch.filter(all_filenames, '%s.GEOTIFF' % matching_exp)
                if not matches:
                    matches = fnmatch.filter(all_filenames, '%s.TIF' % matching_exp)
                if not matches:
                    raise FileNotFoundError("Could not find a file that matches the expression '%s'." % matching_exp)

                return matches[0]

            self.filename_data = xml.find("product/image/%s/name" % lbl).text
            self.scene_basename = self.filename_data.split('-SPECTRAL_IMAGE')[0]
            self.filename_quicklook = xml.find("product/quicklook/%s/name" % lbl).text
            self.filename_deadpixelmap = get_filename('*QL_PIXELMASK_%s' % self.detector_name)
            self.filename_mask_landwater = get_filename('*QL_QUALITY_CLASSES')
            self.filename_mask_snow = get_filename('*QL_QUALITY_SNOW')
            self.filename_mask_cloudshadow = get_filename('*QL_QUALITY_CLOUDSHADOW')
            self.filename_mask_clouds = get_filename('*-QL_QUALITY_CLOUD')
            self.filename_mask_haze = get_filename('*QL_QUALITY_HAZE')
            self.filename_mask_cirrus = get_filename('*QL_QUALITY_CIRRUS')

            # FIXME combine different cloud masks?
            # TODO: Add test flags layer.

            # read some basic information concerning the detector
            self.nrows = int(xml.find("product/image/%s/dimension/rows" % lbl).text)
            self.ncols = int(xml.find("product/image/%s/dimension/columns" % lbl).text)
            self.unitcode = 'DN'
            self.unit = ''

            # Read image coordinates
            # FIXME why do we have the same corner coordinates for VNIR and SWIR?
            points = xml.findall("base/spatialCoverage/boundingPolygon/point")
            coords_xy = {p.find('frame').text: (float(p.find('longitude').text),
                                                float(p.find('latitude').text))
                         for p in points}
            coord_tags = ['upper_left', 'upper_right', 'lower_left', 'lower_right']
            self.lon_UL_UR_LL_LR = [coords_xy[ct][0] for ct in coord_tags]
            self.lat_UL_UR_LL_LR = [coords_xy[ct][1] for ct in coord_tags]

            # read the band related information: wavelength, fwhm
            self.nwvl = int(xml.find("product/image/%s/channels" % lbl).text)
            # FIXME hardcoded - DLR does not provide any smile information
            #   => smile coefficients are set to zero until we have valid ones
            self.nsmile_coef = 5
            self.smile_coef = np.zeros((self.nwvl, self.nsmile_coef), dtype=np.float)

            startband = 0 if self.detector_name == 'VNIR' else int(xml.find("product/image/vnir/channels").text)
            subset = slice(startband, startband + self.nwvl)
            bi = "specific/bandCharacterisation/bandID/"
            self.wvl_center = np.array([float(ele.text) for ele in xml.findall(bi + "wavelengthCenterOfBand")[subset]])
            self.fwhm = np.array([float(ele.text) for ele in xml.findall(bi + "FWHMOfBand")[subset]])
            self.gains = np.array([float(ele.text) for ele in xml.findall(bi + "GainOfBand")[subset]])
            self.offsets = np.array([float(ele.text) for ele in xml.findall(bi + "OffsetOfBand")[subset]])

            # read preview bands
            wvl_red = float(xml.find("product/image/%s/qlChannels/red" % lbl).text)
            wvl_green = float(xml.find("product/image/%s/qlChannels/green" % lbl).text)
            wvl_blue = float(xml.find("product/image/%s/qlChannels/blue" % lbl).text)
            self.preview_wvls = [wvl_red, wvl_green, wvl_blue]
            self.preview_bands = np.array([np.argmin(np.abs(self.wvl_center - wvl)) for wvl in self.preview_wvls])

            # read RPC coefficients
            for bID in xml.findall("product/navigation/RPC/bandID")[subset]:
                bN = 'band_%d' % np.int64(bID.attrib['number'])

                keys2combine = ('row_num', 'row_den', 'col_num', 'col_den')

                tmp = OrderedDict([(ele.tag.lower(), np.float(ele.text)) for ele in bID.findall('./')])
                self.rpc_coeffs[bN] = {k: v for k, v in tmp.items() if not k.startswith(keys2combine)}

                for n in keys2combine:
                    self.rpc_coeffs[bN]['%s_coeffs' % n.lower()] = \
                        np.array([v for k, v in tmp.items() if k.startswith(n)])

        else:
            lbl = self.detector_label
            self.logger.info("Reading metadata for %s detector..." % self.detector_name)

            # read data filenames
            self.filename_data = xml.findall("ProductComponent/%s/Data/Filename" % lbl)[0].text
            self.scene_basename = os.path.splitext(self.filename_data)[0]
            self.filename_deadpixelmap = xml.findall("ProductComponent/%s/Sensor/DeadPixel/Filename" % lbl)[0].text
            self.filename_quicklook = xml.findall("ProductComponent/%s/Preview/Filename" % lbl)[0].text
            self.filename_mask_clouds = xml.findall("ProductComponent/%s/Data/CloudMaskMap/Filename" % lbl)[0].text

            # read preview bands
            self.preview_bands = np.zeros(3, dtype=np.int)
            self.preview_bands[0] = np.int(xml.findall("ProductComponent/%s/Preview/Bands/Red" % lbl)[0].text)
            self.preview_bands[1] = np.int(xml.findall("ProductComponent/%s/Preview/Bands/Green" % lbl)[0].text)
            self.preview_bands[2] = np.int(xml.findall("ProductComponent/%s/Preview/Bands/Blue" % lbl)[0].text)

            # read some basic information concerning the detector
            self.nrows = np.int(xml.findall("ProductComponent/%s/Data/Size/NRows" % lbl)[0].text)
            self.ncols = np.int(xml.findall("ProductComponent/%s/Data/Size/NCols" % lbl)[0].text)
            self.unitcode = xml.findall("ProductComponent/%s/Data/Type/UnitCode" % lbl)[0].text
            self.unit = xml.findall("ProductComponent/%s/Data/Type/Unit" % lbl)[0].text

            # Read image coordinates
            scene_corner_coordinates = xml.findall("ProductComponent/%s/Data/SceneInformation/"
                                                   "SceneCornerCoordinates" % lbl)
            self.lat_UL_UR_LL_LR = [
                np.float(scene_corner_coordinates[0].findall("Latitude")[0].text),
                np.float(scene_corner_coordinates[1].findall("Latitude")[0].text),
                np.float(scene_corner_coordinates[2].findall("Latitude")[0].text),
                np.float(scene_corner_coordinates[3].findall("Latitude")[0].text)
            ]
            self.lon_UL_UR_LL_LR = [
                np.float(scene_corner_coordinates[0].findall("Longitude")[0].text),
                np.float(scene_corner_coordinates[1].findall("Longitude")[0].text),
                np.float(scene_corner_coordinates[2].findall("Longitude")[0].text),
                np.float(scene_corner_coordinates[3].findall("Longitude")[0].text)
            ]

            # read the band related information: wavelength, fwhm
            self.nwvl = np.int(xml.findall("ProductComponent/%s/Data/BandInformationList/NumberOfBands" % lbl)[0].text)
            self.nsmile_coef = np.int(xml.findall(
                "ProductComponent/%s/Data/BandInformationList/SmileInformation/NumberOfCoefficients" % lbl)[0].text)
            self.fwhm = np.zeros(self.nwvl, dtype=np.float)
            self.wvl_center = np.zeros(self.nwvl, dtype=np.float)
            self.smile_coef = np.zeros((self.nwvl, self.nsmile_coef), dtype=np.float)
            self.l_min = np.zeros(self.nwvl, dtype=np.float)
            self.l_max = np.zeros(self.nwvl, dtype=np.float)
            band_informations = xml.findall("ProductComponent/%s/Data/BandInformationList/BandInformation" % lbl)
            for bi in band_informations:
                k = np.int64(bi.attrib['Id']) - 1
                self.wvl_center[k] = np.float(bi.findall("CenterWavelength")[0].text)
                self.fwhm[k] = np.float(bi.findall("FullWidthHalfMaximum")[0].text)
                self.l_min[k] = np.float(bi.findall("L_min")[0].text)
                self.l_max[k] = np.float(bi.findall("L_max")[0].text)
                scl = bi.findall("Smile/Coefficient")
                for sc in scl:
                    self.smile_coef[k, np.int64(sc.attrib['exponent'])] = np.float(sc.text)

            self.lats = self.interpolate_corners(*self.lat_UL_UR_LL_LR, self.ncols, self.nrows)
            self.lons = self.interpolate_corners(*self.lon_UL_UR_LL_LR, self.ncols, self.nrows)
            self.preview_wvls = np.array([self.wvl_center[i] for i in self.preview_bands])

        # compute metadata derived from read data
        self.smile = self.calc_smile()
        self.srf = SRF.from_cwl_fwhm(self.wvl_center, self.fwhm)
        self.solar_irrad = self.calc_solar_irradiance_CWL_FWHM_per_band()
        self.ll_mapPoly = get_footprint_polygon(tuple(zip(self.lon_UL_UR_LL_LR,
                                                          self.lat_UL_UR_LL_LR)), fix_invalid=True)
        from ...processors.spatial_transform import get_UTMEPSG_from_LonLat_cornersXY
        self.epsg_ortho = get_UTMEPSG_from_LonLat_cornersXY(lons=self.lon_UL_UR_LL_LR, lats=self.lat_UL_UR_LL_LR)

    def calc_smile(self):
        """Compute smile for each EnMAP column.

        The sum in (1) is expressed as inner product of two arrays with inner dimension as the polynomial smile
        coefficients shape = (ncols, nsmile_coef) of polynomial x

        :return:
        """
        # smile(icol, iwvl) = sum_(p=0)^(nsmile_coef-1) smile_coef[iwvl, p] * icol**p (1)
        return np.inner(
            np.array([[icol ** p for p in np.arange(self.nsmile_coef)] for icol in np.arange(self.ncols)]),
            self.smile_coef  # shape = (nwvl, nsmile_coef)
        )  # shape = (ncols, nwvl)

    def calc_snr_from_radiance(self, rad_data: Union[GeoArray, np.ndarray], dir_snr_models: str):
        """Compute EnMAP SNR from radiance data for the given detector.

        SNR equation:    SNR = p0 + p1 * LTOA + p2 * LTOA ^ 2   [W/(m^2 sr nm)].
        NOTE:   The SNR model files (SNR_D1.bsq/SNR_D2.bsq) contain polynomial coefficients needed to compute SNR.
                SNR_D1.bsq: SNR model for EnMAP SWIR detector (contains high gain and low gain model coefficients)
                            - 1000 columns (for 1000 EnMAP columns)
                            - 88 bands for 88 EnMAP VNIR bands
                            - 7 lines:  - 3 lines: high gain coefficients
                                        - 3 lines: low gain coefficients
                                        - 1 line: threshold needed to decide about high gain or low gain
                SNR_D2.bsq: SNR model for EnMAP SWIR detector
                            - 1000 columns (for 1000 EnMAP columns)
                            - x bands for x EnMAP SWIR bands
                            - 3 lines for 3 coefficients

        :param rad_data:        image radiance data of EnMAP_Detector_SensorGeo
        :param dir_snr_models:  root directory where SNR model data is stored (must contain SNR_D1.bsq/SNR_D2.bsq)
        """
        path_snr_model = os.path.join(dir_snr_models, "SNR_D1.bsq" if self.detector_name == 'VNIR' else "SNR_D2.bsq")
        rad_data = np.array(rad_data)

        assert self.unitcode == 'TOARad'
        self.logger.info("Computing SNR for %s using %s" % (self.detector_name, path_snr_model))

        if self.detector_name == 'VNIR':
            gA = GeoArray(path_snr_model)
            coeffs_highgain = gA[0:3, :, :]  # [3 x ncols x nbands]
            coeffs_lowgain = gA[3:6, :, :]  # [3 x ncols x nbands]
            gain_threshold = np.squeeze(gA[6, :, :])

            self.snr = np.zeros(rad_data.shape)
            for irow in range(self.nrows):
                highgain_mask = rad_data[irow, :, :] < gain_threshold  # a single row
                rad_highgain = rad_data[irow, highgain_mask]
                self.snr[irow, :, :][highgain_mask] = \
                    coeffs_highgain[0, highgain_mask] + \
                    coeffs_highgain[1, highgain_mask] * rad_highgain + \
                    coeffs_highgain[2, highgain_mask] * rad_highgain ** 2

                lowgain_mask = rad_data[irow, :, :] >= gain_threshold
                rad_lowgain = rad_data[irow, lowgain_mask]
                self.snr[irow, :, :][lowgain_mask] = \
                    coeffs_lowgain[0, lowgain_mask] + \
                    coeffs_lowgain[1, lowgain_mask] * rad_lowgain + \
                    coeffs_lowgain[2, lowgain_mask] * rad_lowgain ** 2

        else:
            gA = GeoArray(path_snr_model)

            # some SWIR bands may be missing -> find indices of bands we need for SNR computation
            cwls_gA = gA.metadata.band_meta['wavelength']
            cwls_needed = self.wvl_center

            idxs_needed = [np.argmin(np.abs(cwls_gA - cwl)) for cwl in cwls_needed]
            if not len(set(idxs_needed)) == len(idxs_needed) or len(idxs_needed) != self.nwvl:
                raise RuntimeError('Unclear band allocation during SWIR SNR computation.')

            coeffs = gA[:, :, idxs_needed]  # [3 x ncols x nbands]
            self.snr = coeffs[0, :, :] + coeffs[1, :, :] * rad_data[:, :, :] + coeffs[2, :, :] * rad_data[:, :, :] ** 2

    @staticmethod
    def interpolate_corners(ul: float, ur: float, ll: float, lr: float, nx: int, ny: int):
        """Compute interpolated field from corner values of a scalar field given at: ul, ur, ll, lr.

        :param ul:  tbd
        :param ur:  tbd
        :param ll:  tbd
        :param lr:  tbd
        :param nx: final shape (x-axis direction)
        :param ny: final shape (y-axis direction)
        """
        # FIXME this method must later be replaced by the geolayer provided by the ground segment
        #       => a linear interpolation between the EnMAP corner coordinates is NOT sufficient for modelling the
        #          geometry of VNIR and SWIR
        #       - especially at off-nadir acquisitions and with some terrain present, a linear interpolation leads
        #         to large deviations (> 180 m y-coordinate offset for the EnPT test dataset)

        # TODO ensure that lons/lats represent UL coordinates not pixel coordinate centers (as given by Karl / DLR(?))

        corner_coords = np.array([[ul, ur],
                                  [ll, lr]])
        rowpos, colpos = [0, 1], [0, 1]

        from scipy.interpolate import RegularGridInterpolator
        rgi = RegularGridInterpolator([rowpos, colpos], corner_coords, method='linear')
        out_rows_grid, out_cols_grid = np.meshgrid(np.linspace(0, 1, ny),
                                                   np.linspace(0, 1, nx),
                                                   indexing='ij')

        coords = rgi(np.dstack([out_rows_grid, out_cols_grid]))

        return coords

    def compute_geolayer_for_cube(self):
        self.logger.info('Computing %s geolayer...' % self.detector_name)

        lons, lats = RPC_3D_Geolayer_Generator(rpc_coeffs_per_band=self.rpc_coeffs,
                                               dem=self.cfg.path_dem,
                                               enmapIm_cornerCoords=tuple(zip(self.lon_UL_UR_LL_LR,
                                                                              self.lat_UL_UR_LL_LR)),
                                               enmapIm_dims_sensorgeo=(self.nrows, self.ncols),
                                               CPUs=self.cfg.CPUs)\
            .compute_geolayer()

        return lons, lats

    def calc_solar_irradiance_CWL_FWHM_per_band(self) -> np.array:
        from ...io.reader import Solar_Irradiance_reader

        self.logger.debug('Calculating solar irradiance...')

        sol_irr = Solar_Irradiance_reader(path_solar_irr_model=self.cfg.path_solar_irr, wvl_min_nm=350, wvl_max_nm=2500)

        irr_bands = []
        for band in self.srf.bands:
            WVL_band = self.srf.srfs_wvl if self.srf.wvl_unit == 'nanometers' else self.srf.srfs_wvl * 1000
            RSP_band = self.srf.srfs_norm01[band]
            sol_irr_at_WVL = np.interp(WVL_band, sol_irr[:, 0], sol_irr[:, 1], left=0, right=0)

            irr_bands.append(np.round(np.sum(sol_irr_at_WVL * RSP_band) / np.sum(RSP_band), 2))

        return np.array(irr_bands)


class EnMAP_Metadata_L1B_SensorGeo(object):
    """EnMAP Metadata class for the metadata of the complete EnMAP L1B product in sensor geometry incl. VNIR and SWIR.

    Attributes:
        - logger(logging.Logger):  None or logging instance
        - observation_datetime(datetime.datetime):  datetime of observation time
        - geom_view_zenith: viewing zenith angle
        - geom_view_azimuth: viewing azimuth angle
        - geom_sun_zenith: sun zenith angle
        - geom_sun_azimuth: sun azimuth angle
        - mu_sun: needed by SICOR for TOARad > TOARef conversion
        - vnir(EnMAP_Metadata_VNIR_SensorGeo)
        - swir(EnMAP_Metadata_SWIR_SensorGeo)
        - detector_attrNames: attribute names of the detector objects
    """

    def __init__(self, path_metaxml, config: EnPTConfig, logger=None):
        """Get a metadata object instance for the given EnMAP L1B product in sensor geometry.

        :param path_metaxml:    file path of the EnMAP L1B metadata XML file
        :para, config:          EnPT configuration object
        :param logger:          instance of logging.logger or subclassed
        """
        self.cfg = config
        self.logger = logger or logging.getLogger()
        self.path_xml = path_metaxml
        self.rootdir = os.path.dirname(path_metaxml)

        # defaults - Common
        self.proc_level: Optional[str] = None   # Dataset processing level
        self.observation_datetime: Optional[datetime] = None  # Date and Time of image observation
        self.geom_view_zenith: Optional[float] = None  # viewing zenith angle
        self.geom_view_azimuth: Optional[float] = None  # viewing azimuth angle
        self.geom_sun_zenith: Optional[float] = None  # sun zenith angle
        self.geom_sun_azimuth: Optional[float] = None   # sun azimuth angle
        self.mu_sun: Optional[float] = None   # needed by SICOR for TOARad > TOARef conversion
        self.earthSunDist: Optional[float] = None  # earth-sun distance
        self.aot: Optional[float] = None  # scene aerosol optical thickness
        self.water_vapour: Optional[float] = None  # scene water vapour [cm]
        self.vnir: Optional[EnMAP_Metadata_L1B_Detector_SensorGeo] = None  # metadata of VNIR only
        self.swir: Optional[EnMAP_Metadata_L1B_Detector_SensorGeo] = None  # metadata of SWIR only
        self.detector_attrNames: list = ['vnir', 'swir']  # attribute names of the detector objects
        self.filename_metaxml: Optional[str] = None  # filename of XML metadata file

        self._scene_basename: Optional[str] = None  # basename of the EnMAP image

    @property
    def scene_basename(self):
        if self.vnir:
            self._scene_basename = self.vnir.scene_basename

        return self._scene_basename

    # Read common metadata method
    def read_common_meta(self, path_xml):
        """Read the common metadata, principally stored in General Info
        - the acquisition time
        - the geometrical observation and illumination
        :param path_xml: path to the main xml file
        :return: None
        """

        # load the metadata xml file
        xml = ElementTree.parse(path_xml).getroot()

        self.filename_metaxml = os.path.basename(path_xml)

        if not self.cfg.is_dummy_dataformat:
            # read processing level
            self.proc_level = xml.find("base/level").text
            if self.proc_level != 'L1B':
                raise RuntimeError(self.proc_level, "Unexpected input data processing level. Expected 'L1B'.")

            # read the acquisition time
            self.observation_datetime = \
                datetime.strptime(xml.find("base/temporalCoverage/startTime").text, '%Y-%m-%dT%H:%M:%S.%fZ')

            # get the distance earth sun from the acquisition date
            self.earthSunDist = self.get_earth_sun_distance(self.observation_datetime)

            # read Geometry (observation/illumination) angle
            # NOTE: EnMAP metadata provide also the angles for the image corners
            #       -> would allow even more precise computation (e.g., specific/sunElevationAngle/upper_left)
            # NOTE: alongOffNadirAngle is always near 0 and therefore ignored here (not relevant for AC)
            # FIXME VZA may be negative in DLR L1B data -> correct to always use the absolute value for SICOR?
            self.geom_view_zenith = np.abs(np.float(xml.find("specific/acrossOffNadirAngle/center").text))
            # FIXME correct to directly use sceneAzimuthAngle (14.24 (DLR) vs. 101.1 (AlpineTest)
            self.geom_view_azimuth = np.float(xml.find("specific/sceneAzimuthAngle/center").text)
            self.geom_sun_zenith = 90 - np.float(xml.find("specific/sunElevationAngle/center").text)
            self.geom_sun_azimuth = np.float(xml.find("specific/sunAzimuthAngle/center").text)
            self.mu_sun = np.cos(np.deg2rad(self.geom_sun_zenith))
            self.aot = np.float(xml.find("specific/qualityFlag/sceneAOT").text) / 1000  # scale factor is 1000
            self.water_vapour = np.float(xml.find("specific/qualityFlag/sceneWV").text) / 1000  # scale factor is 1000
        else:
            # read the acquisition time
            self.observation_datetime = \
                datetime.strptime(xml.findall("GeneralInfo/ProductInfo/ProductStartTime")[0].text,
                                  '%Y-%m-%dT%H:%M:%S.%fZ')

            # get the distance earth sun from the acquisition date
            self.earthSunDist = self.get_earth_sun_distance(self.observation_datetime)

            # read Geometry (observation/illumination) angle
            self.geom_view_zenith = np.float(xml.findall("GeneralInfo/Geometry/Observation/ZenithAngle")[0].text)
            self.geom_view_azimuth = np.float(xml.findall("GeneralInfo/Geometry/Observation/AzimuthAngle")[0].text)
            self.geom_sun_zenith = np.float(xml.findall("GeneralInfo/Geometry/Illumination/ZenithAngle")[0].text)
            self.geom_sun_azimuth = np.float(xml.findall("GeneralInfo/Geometry/Illumination/AzimuthAngle")[0].text)
            self.mu_sun = np.cos(np.deg2rad(self.geom_sun_zenith))

    def get_earth_sun_distance(self, acqDate: datetime):
        """Get earth sun distance (requires file of pre calculated earth sun distance per day)

        :param acqDate:
        """
        if not os.path.exists(self.cfg.path_earthSunDist):
            self.logger.warning("\n\t WARNING: Earth Sun Distance is assumed to be "
                                "1.0 because no database can be found at %s.""" % self.cfg.path_earthSunDist)
            return 1.0
        if not acqDate:
            self.logger.warning("\n\t WARNING: Earth Sun Distance is assumed to be 1.0 because "
                                "acquisition date could not be read from metadata.")
            return 1.0

        with open(self.cfg.path_earthSunDist, "r") as EA_dist_f:
            EA_dist_dict = {}
            for line in EA_dist_f:
                date, EA = [item.strip() for item in line.split(",")]
                EA_dist_dict[date] = EA

        return float(EA_dist_dict[acqDate.strftime('%Y-%m-%d')])

    def read_metadata(self):
        """
        Read the metadata of the entire EnMAP L1B product in sensor geometry

        :return: None
        """

        # first read common metadata
        self.read_common_meta(self.path_xml)

        # define and read the VNIR metadata
        self.vnir = EnMAP_Metadata_L1B_Detector_SensorGeo('VNIR', config=self.cfg, logger=self.logger)
        self.vnir.read_metadata(self.path_xml)

        # define and read the SWIR metadata
        self.swir = EnMAP_Metadata_L1B_Detector_SensorGeo('SWIR', config=self.cfg, logger=self.logger)
        self.swir.read_metadata(self.path_xml)

    def to_XML(self) -> str:
        """
        Generate an XML metadata string from the L1B metadata.
        """
        from . import L1B_product_props, L1B_product_props_DLR
        xml = ElementTree.parse(self.path_xml).getroot()

        if not self.cfg.is_dummy_dataformat:
            for detName, detMeta in zip(['VNIR', 'SWIR'], [self.vnir, self.swir]):
                lbl = L1B_product_props_DLR['xml_detector_label'][detName]
                xml.find("product/image/%s/dimension/rows" % lbl).text = str(detMeta.nrows)
                xml.find("product/image/%s/dimension/columns" % lbl).text = str(detMeta.ncols)
                xml.find("product/quicklook/%s/dimension/rows" % lbl).text = str(detMeta.nrows)
                xml.find("product/quicklook/%s/dimension/columns" % lbl).text = str(detMeta.ncols)

        else:
            for detName, detMeta in zip(['VNIR', 'SWIR'], [self.vnir, self.swir]):
                lbl = L1B_product_props['xml_detector_label'][detName]
                xml.find("ProductComponent/%s/Data/Size/NRows" % lbl).text = str(detMeta.nrows)
                xml.find("ProductComponent/%s/Data/Type/UnitCode" % lbl).text = detMeta.unitcode
                xml.find("ProductComponent/%s/Data/Type/Unit" % lbl).text = detMeta.unit

        xml_string = ElementTree.tostring(xml, encoding='unicode', pretty_print=True)

        return xml_string



# -*- coding: utf-8 -*-

# EnPT, EnMAP Processing Tool - A Python package for pre-processing of EnMAP Level-1B data
#
# Copyright (C) 2019  Karl Segl (GFZ Potsdam, segl@gfz-potsdam.de), Daniel Scheffler
# (GFZ Potsdam, danschef@gfz-potsdam.de), Niklas Bohn (GFZ Potsdam, nbohn@gfz-potsdam.de),
# Stéphane Guillaso (GFZ Potsdam, stephane.guillaso@gfz-potsdam.de)
#
# This software was developed within the context of the EnMAP project supported
# by the DLR Space Administration with funds of the German Federal Ministry of
# Economic Affairs and Energy (on the basis of a decision by the German Bundestag:
# 50 EE 1529) and contributions from DLR, GFZ and OHB System AG.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version. Please note the following exception: `EnPT` depends on tqdm, which
# is distributed under the Mozilla Public Licence (MPL) v2.0 except for the files
# "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""EnPT process controller module."""

import os
import tempfile
import zipfile
import shutil
import weakref
import warnings
import pickle
from typing import Optional

from ..options.config import EnPTConfig
from ..io.reader import L1B_Reader
from ..model.images import EnMAPL1Product_SensorGeo, EnMAPL2Product_MapGeo  # noqa: F401

__author__ = 'Daniel Scheffler'


class EnPT_Controller(object):
    """Class of EnPT process controller."""

    def __init__(self, config: EnPTConfig = None, **config_kwargs: dict):
        """Initialize the Process Controller.

        :param config:          an instance of the EnPTConfig class (overrides config_kwargs)
        :param config_kwargs:   configuration parameters to be passed to EnPTConfig class
        """
        self.cfg: EnPTConfig = config or EnPTConfig(**config_kwargs)

        # generate temporary directory (must be deleted at the end of the pipeline)
        self.tempDir = tempfile.mkdtemp()

        # setup a finalizer that destroys remaining data (directories, etc.) in case of unexpected exit
        self._finalizer = weakref.finalize(self, self._cleanup, self.tempDir,
                                           warn_message="Implicitly cleaning up {!r}".format(self))

        # defaults
        self.L1_obj: Optional[EnMAPL1Product_SensorGeo] = None
        self.L2_obj: Optional[EnMAPL2Product_MapGeo] = None

    def extract_zip_archive(self, path_zipfile: str, subdir: str = '') -> str:
        """Extract the given EnMAP image zip archive and return the L1B image root directory path.

        :param path_zipfile:    /path/to/zipfile.zip
        :param subdir:          subdirectory name to be created within temporary directory
        :return:                /tmp/tmpk2qp0yri/rootdir/
        """
        outdir = os.path.join(self.tempDir, subdir) if not self.cfg.is_dummy_dataformat else self.tempDir

        with zipfile.ZipFile(path_zipfile, "r") as zf:
            zf.extractall(outdir)

        if not os.path.isdir(outdir):
            raise NotADirectoryError(outdir)

        if not self.cfg.is_dummy_dataformat:
            return outdir
        else:
            return os.path.join(self.tempDir, os.path.basename(path_zipfile).split('.zip')[0])

    def read_L1B_data(self) -> None:
        """Read the provider L1B data given in config and return an EnMAP image object."""
        path_enmap_image = self.cfg.path_l1b_enmap_image
        path_enmap_image_gapfill = self.cfg.path_l1b_enmap_image_gapfill

        # input validation
        if not os.path.isdir(path_enmap_image) and \
           not (os.path.exists(path_enmap_image) and path_enmap_image.endswith('.zip')):
            raise ValueError("The parameter 'path_enmap_image' must be a directory or the path to an existing zip "
                             "archive. Received %s." % path_enmap_image)

        # extract L1B image archive if needed
        if path_enmap_image.endswith('.zip'):
            path_enmap_image = self.extract_zip_archive(path_enmap_image, subdir='image_main')

        # extract L1B gap fill image archive if needed
        if path_enmap_image_gapfill and path_enmap_image_gapfill.endswith('.zip'):
            path_enmap_image_gapfill = self.extract_zip_archive(path_enmap_image_gapfill, subdir='image_gapfill')

        # run the reader
        RD = L1B_Reader(config=self.cfg)
        self.L1_obj = RD.read_inputdata(root_dir_main=path_enmap_image, root_dir_ext=path_enmap_image_gapfill,
                                        n_line_ext=self.cfg.n_lines_to_append)

    def run_toaRad2toaRef(self):
        """Run conversion from TOA radiance to TOA reflectance."""
        # get a new instance of radiometric transformer
        from ..processors.radiometric_transform import Radiometric_Transformer
        RT = Radiometric_Transformer(self.cfg)

        # run transformation to TOARef
        self.L1_obj = RT.transform_TOARad2TOARef(self.L1_obj)

    def run_dem_processor(self):
        self.L1_obj.get_preprocessed_dem()

    def run_geometry_processor(self):
        pass

    def run_atmospheric_correction(self):
        """Run atmospheric correction only."""
        self.L1_obj.run_AC()

    def run_orthorectification(self):
        """Run orthorectification to transform L1 sensor geometry image to L2 map geometry."""
        # run transformation from sensor to map geometry
        self.L2_obj = EnMAPL2Product_MapGeo.from_L1B_sensorgeo(config=self.cfg, enmap_ImageL1=self.L1_obj)

    def write_output(self):
        if self.cfg.output_dir:
            if self.L2_obj is not None:
                self.L2_obj.save(self.cfg.output_dir)
            else:
                self.L1_obj.save(self.cfg.output_dir)

    def run_all_processors(self):
        """Run all processors at once."""
        if os.getenv('IS_ENPT_GUI_TEST') != "1":
            try:
                if self.cfg.log_level == 'DEBUG':
                    self._print_received_configuration()

                self.read_L1B_data()
                if self.cfg.run_deadpix_P:
                    self.L1_obj.correct_dead_pixels()
                # self.run_toaRad2toaRef()  # this is only needed for geometry processor but AC expects radiance
                self.run_dem_processor()
                if self.cfg.enable_ac:
                    self.run_atmospheric_correction()

                    # re-apply dead pixel correction
                    self.L1_obj.logger.info(
                        'Re-applying dead pixel correction to correct fof spectral spikes due to fringe effect.')
                    self.L1_obj.correct_dead_pixels()
                else:
                    self.L1_obj.logger.info('Skipping atmospheric correction as configured and '
                                            'computing top-of-atmosphere reflectance instead.')
                    self.run_toaRad2toaRef()
                self.run_geometry_processor()
                self.run_orthorectification()
                self.write_output()
            finally:
                self.cleanup()

        else:
            self._print_received_configuration()

            if not os.path.isdir(self.cfg.output_dir):
                raise NotADirectoryError(self.cfg.output_dir)

            with open(os.path.join(self.cfg.output_dir, 'received_args_kwargs.pkl'), 'wb') as outF:
                pickle.dump(
                    dict(
                        json_config=self.cfg.json_config,
                        kwargs=self.cfg.kwargs
                    ),
                    outF)

    def _print_received_configuration(self):
        print('EnPT Controller received the following configuration:')
        print(repr(self.cfg))

    @classmethod
    def _cleanup(cls, tempDir, warn_message):
        """Clean up implicitly (not to be called directly)."""
        if tempDir:
            shutil.rmtree(tempDir)
        warnings.warn(warn_message, ResourceWarning)

    def cleanup(self):
        """Clean up (to be called directly)."""
        if self._finalizer.detach():
            if self.tempDir:
                shutil.rmtree(self.tempDir)



# -*- coding: utf-8 -*-

# EnPT, EnMAP Processing Tool - A Python package for pre-processing of EnMAP Level-1B data
#
# Copyright (C) 2019  Karl Segl (GFZ Potsdam, segl@gfz-potsdam.de), Daniel Scheffler
# (GFZ Potsdam, danschef@gfz-potsdam.de), Niklas Bohn (GFZ Potsdam, nbohn@gfz-potsdam.de),
# Stéphane Guillaso (GFZ Potsdam, stephane.guillaso@gfz-potsdam.de)
#
# This software was developed within the context of the EnMAP project supported
# by the DLR Space Administration with funds of the German Federal Ministry of
# Economic Affairs and Energy (on the basis of a decision by the German Bundestag:
# 50 EE 1529) and contributions from DLR, GFZ and OHB System AG.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version. Please note the following exception: `EnPT` depends on tqdm, which
# is distributed under the Mozilla Public Licence (MPL) v2.0 except for the files
# "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""EnPT EnMAP objects in map geometry."""

import logging
from types import SimpleNamespace
from typing import Tuple, Optional  # noqa: F401
from os import path, makedirs

from geoarray import GeoArray, NoDataMask

from .image_baseclasses import _EnMAP_Image
from .images_sensorgeo import EnMAPL1Product_SensorGeo
from ...utils.logging import EnPT_Logger
from ...model.metadata import EnMAP_Metadata_L2A_MapGeo  # noqa: F401  # only used for type hint
from ...options.config import EnPTConfig

__author__ = ['Daniel Scheffler', 'Stéphane Guillaso', 'André Hollstein']


class EnMAP_Detector_MapGeo(_EnMAP_Image):
    """Base class representing a single detector of an EnMAP image (as map geometry).

    NOTE:
        - Inherits all attributes from _EnMAP_Image class.
        - All functionality that VNIR and SWIR detectors (map geometry) have in common is to be implemented here.
        - All EnMAP image subclasses representing a specific EnMAP detector (sensor geometry) should inherit from
          _EnMAP_Detector_SensorGeo.

    Attributes:
        - to be listed here. Check help(_EnMAP_Detector_SensorGeo) in the meanwhile!

    """

    def __init__(self, detector_name: str, logger=None):
        """Get an instance of _EnMAP_Detector_MapGeo.

        :param detector_name:   'VNIR' or 'SWIR'
        :param logger:
        """
        self.detector_name = detector_name
        self.logger = logger or logging.getLogger()

        # private attributes
        self._mask_nodata = None

        # get all attributes of base class "_EnMAP_Image"
        super(EnMAP_Detector_MapGeo, self).__init__()

    @property
    def mask_nodata(self) -> GeoArray:
        """Return the no data mask.

        Bundled with all the corresponding metadata.

        For usage instructions and a list of attributes refer to help(self.data).
        self.mask_nodata works in the same way.

        :return: instance of geoarray.NoDataMask
        """
        if self._mask_nodata is None and isinstance(self.data, GeoArray):
            self.logger.info('Calculating nodata mask...')
            self._mask_nodata = self.data.mask_nodata  # calculates mask nodata if not already present

        return self._mask_nodata

    @mask_nodata.setter
    def mask_nodata(self, *geoArr_initArgs):
        self._mask_nodata = self._get_geoarray_with_datalike_geometry(geoArr_initArgs, 'mask_nodata',
                                                                      nodataVal=False, specialclass=NoDataMask)

    @mask_nodata.deleter
    def mask_nodata(self):
        self._mask_nodata = None

    def calc_mask_nodata(self, fromBand=None, overwrite=False) -> GeoArray:
        """Calculate a no data mask with (values: 0=nodata; 1=data).

        :param fromBand:  <int> index of the band to be used (if None, all bands are used)
        :param overwrite: <bool> whether to overwrite existing nodata mask that has already been calculated
        :return:
        """
        self.logger.info('Calculating nodata mask...')

        if self._mask_nodata is None or overwrite:
            self.data.calc_mask_nodata(fromBand=fromBand, overwrite=overwrite)
            self.mask_nodata = self.data.mask_nodata
            return self.mask_nodata


class EnMAPL2Product_MapGeo(_EnMAP_Image):
    """Class for EnPT Level-2 EnMAP object in map geometry.

    Attributes:
        - logger:
            - logging.Logger instance or subclass instance
        - paths:
            - paths belonging to the EnMAP product
        - meta:
            - instance of EnMAP_Metadata_SensorGeo class
    """
    def __init__(self, config: EnPTConfig, logger=None):
        # protected attributes
        self._logger = None

        # populate attributes
        self.cfg = config
        if logger:
            self.logger = logger

        self.meta: Optional[EnMAP_Metadata_L2A_MapGeo] = None
        self.paths: Optional[SimpleNamespace] = None

        super(EnMAPL2Product_MapGeo, self).__init__()

    @property
    def logger(self) -> EnPT_Logger:
        """Get a an instance of enpt.utils.logging.EnPT_Logger.

        NOTE:
            - The logging level will be set according to the user inputs of EnPT.
            - The path of the log file is directly derived from the attributes of the _EnMAP_Image instance.

        Usage:
            - get the logger:
                logger = self.logger
            - set the logger
                self.logger = logging.getLogger()  # NOTE: only instances of logging.Logger are allowed here
            - delete the logger:
                del self.logger  # or "self.logger = None"

        :return: EnPT_Logger instance
        """
        if self._logger and self._logger.handlers[:]:
            return self._logger
        else:
            basename = path.splitext(path.basename(self.cfg.path_l1b_enmap_image))[0]
            path_logfile = path.join(self.cfg.output_dir, basename + '.log') \
                if self.cfg.create_logfile and self.cfg.output_dir else ''
            self._logger = EnPT_Logger('log__' + basename, fmt_suffix=None, path_logfile=path_logfile,
                                       log_level=self.cfg.log_level, append=False)

            return self._logger

    @logger.setter
    def logger(self, logger: logging.Logger):
        assert isinstance(logger, logging.Logger) or logger in ['not set', None], \
            "%s.logger can not be set to %s." % (self.__class__.__name__, logger)

        # save prior logs
        # if logger is None and self._logger is not None:
        #     self.log += self.logger.captured_stream
        self._logger = logger

    @property
    def log(self) -> str:
        """Return a string of all logged messages until now.

        NOTE: self.log can also be set to a string.
        """
        return self.logger.captured_stream

    @log.setter
    def log(self, string: str):
        assert isinstance(string, str), "'log' can only be set to a string. Got %s." % type(string)
        self.logger.captured_stream = string

    @classmethod
    def from_L1B_sensorgeo(cls, config: EnPTConfig, enmap_ImageL1: EnMAPL1Product_SensorGeo):
        from ...processors.orthorectification import Orthorectifier
        L2_obj = Orthorectifier(config=config).run_transformation(enmap_ImageL1=enmap_ImageL1)

        return L2_obj

    def get_paths(self, l2a_outdir: str):
        """
        Get all file paths associated with the current instance of EnMAP_Detector_SensorGeo
        These information are read from the detector_meta.

        :param l2a_outdir:  output directory of EnMAP Level-2A dataset
        :return: paths as SimpleNamespace
        """
        paths = SimpleNamespace()
        paths.root_dir = l2a_outdir
        paths.metaxml = path.join(l2a_outdir, self.meta.filename_metaxml)
        paths.data = path.join(l2a_outdir, self.meta.filename_data)
        paths.mask_landwater = path.join(l2a_outdir, self.meta.filename_mask_landwater)
        paths.mask_clouds = path.join(l2a_outdir, self.meta.filename_mask_clouds)
        paths.mask_cloudshadow = path.join(l2a_outdir, self.meta.filename_mask_cloudshadow)
        paths.mask_haze = path.join(l2a_outdir, self.meta.filename_mask_haze)
        paths.mask_snow = path.join(l2a_outdir, self.meta.filename_mask_snow)
        paths.mask_cirrus = path.join(l2a_outdir, self.meta.filename_mask_cirrus)
        paths.deadpixelmap_vnir = path.join(l2a_outdir, self.meta.filename_deadpixelmap_vnir)
        paths.deadpixelmap_swir = path.join(l2a_outdir, self.meta.filename_deadpixelmap_swir)
        paths.quicklook_vnir = path.join(l2a_outdir, self.meta.filename_quicklook_vnir)
        paths.quicklook_swir = path.join(l2a_outdir, self.meta.filename_quicklook_swir)

        return paths

    def save(self, outdir: str, suffix="") -> str:
        """
        Save the product to disk using almost the same input format
        :param outdir: path to the output directory
        :param suffix: suffix to be appended to the output filename (???)
        :return: root path (root directory) where products were written
        """
        # TODO optionally add more output formats
        product_dir = path.join(path.abspath(outdir),
                                "{name}{suffix}".format(name=self.meta.scene_basename, suffix=suffix))

        self.logger.info("Write product to: %s" % product_dir)
        makedirs(product_dir, exist_ok=True)

        # save raster data
        kwargs_save = dict(fmt='GTiff', creationOptions=["COMPRESS=LZW"])
        outpaths = dict(metaxml=path.join(product_dir, self.meta.filename_metaxml))

        for attrName in ['data', 'mask_landwater', 'mask_clouds', 'mask_cloudshadow', 'mask_haze', 'mask_snow',
                         'mask_cirrus', 'quicklook_vnir', 'quicklook_swir', 'deadpixelmap']:

            if attrName == 'deadpixelmap':
                # TODO VNIR and SWIR must be merged
                self.logger.warning('Currently, L2A dead pixel masks cannot be saved yet.')
                continue

            outpath = path.join(product_dir, getattr(self.meta, 'filename_%s' % attrName))

            attr_gA = \
                self.generate_quicklook(bands2use=self.meta.preview_bands_vnir) if attrName == 'quicklook_vnir' else \
                self.generate_quicklook(bands2use=self.meta.preview_bands_swir) if attrName == 'quicklook_swir' else \
                getattr(self, attrName)

            if attr_gA is not None:
                attr_gA.save(outpath, **kwargs_save)
                outpaths[attrName] = outpath
            else:
                self.logger.warning("The '%s' attribute cannot be saved because it does not exist in the current EnMAP "
                                    "image." % attrName)

        # TODO remove GDAL's *.aux.xml files?

        # save metadata
        self.meta.add_product_fileinformation(filepaths=list(outpaths.values()))
        metadata_string = self.meta.to_XML()

        with open(outpaths['metaxml'], 'w') as metaF:
            self.logger.info("Writing metdata to %s" % outpaths['metaxml'])
            metaF.write(metadata_string)

        self.logger.info("L2A product successfully written!")

        return product_dir



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# EnPT, EnMAP Processing Tool - A Python package for pre-processing of EnMAP Level-1B data
#
# Copyright (C) 2019  Karl Segl (GFZ Potsdam, segl@gfz-potsdam.de), Daniel Scheffler
# (GFZ Potsdam, danschef@gfz-potsdam.de), Niklas Bohn (GFZ Potsdam, nbohn@gfz-potsdam.de),
# Stéphane Guillaso (GFZ Potsdam, stephane.guillaso@gfz-potsdam.de)
#
# This software was developed within the context of the EnMAP project supported
# by the DLR Space Administration with funds of the German Federal Ministry of
# Economic Affairs and Energy (on the basis of a decision by the German Bundestag:
# 50 EE 1529) and contributions from DLR, GFZ and OHB System AG.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version. Please note the following exception: `EnPT` depends on tqdm, which
# is distributed under the Mozilla Public Licence (MPL) v2.0 except for the files
# "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, find_packages

__author__ = 'Daniel Scheffler'


with open('README.rst') as readme_file:
    readme = readme_file.read()

version = {}
with open("enpt/version.py", encoding='utf-8') as version_file:
    exec(version_file.read(), version)

requirements = [  # put package requirements here
    'numpy', 'pandas', 'scipy', 'geoarray>=0.8.11', 'py_tools_ds>=0.14.23', 'arosics>=0.9.2', 'sensormapgeo>=0.2.1',
    'cerberus', 'jsmin', 'matplotlib', 'tqdm', 'utm', 'lxml', 'numpy-indexed', 'mvgavg',
    'sicor @ git+https://gitext.gfz-potsdam.de/EnMAP/sicor.git#egg=sicor'
]

setup_requirements = ['setuptools-git']  # needed for package_data version controlled by GIT

test_requirements = ['coverage', 'nose', 'nose-htmloutput', 'rednose']

setup(
    author="Karl Segl, Daniel Scheffler, Niklas Bohn, Stéphane Guillaso",
    author_email="segl@gfz-potsdam.de, danschef@gfz-potsdam.de, nbohn@gfz-potsdam.de, stephane.guillaso@gfz-potsdam.de",
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="EnMAP Processing Tool",
    keywords=['EnPT', 'EnMAP', 'EnMAP-Box', 'hyperspectral', 'remote sensing', 'satellite', 'processing chain'],
    include_package_data=True,
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme,
    name='enpt',
    package_dir={'enpt': 'enpt'},
    # NOTE: if the 'package_data' files are not under CVS or Subversion version control, we need setuptools-git here,
    #       otherwise they are not included in the PyPi upload content
    package_data={"enpt": ["resources/**/**/*"]},
    packages=find_packages(exclude=['tests*', 'examples*']),  # does not seems to work, therefore use MANIFEST.in
    scripts=['bin/enpt_cli.py'],
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://gitext.gfz-potsdam.de/EnMAP/GFZ_Tools_EnMAP_BOX/EnPT',
    version=version['__version__'],
    zip_safe=False
)



# -*- coding: utf-8 -*-

# EnPT, EnMAP Processing Tool - A Python package for pre-processing of EnMAP Level-1B data
#
# Copyright (C) 2019  Karl Segl (GFZ Potsdam, segl@gfz-potsdam.de), Daniel Scheffler
# (GFZ Potsdam, danschef@gfz-potsdam.de), Niklas Bohn (GFZ Potsdam, nbohn@gfz-potsdam.de),
# Stéphane Guillaso (GFZ Potsdam, stephane.guillaso@gfz-potsdam.de)
#
# This software was developed within the context of the EnMAP project supported
# by the DLR Space Administration with funds of the German Federal Ministry of
# Economic Affairs and Energy (on the basis of a decision by the German Bundestag:
# 50 EE 1529) and contributions from DLR, GFZ and OHB System AG.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version. Please note the following exception: `EnPT` depends on tqdm, which
# is distributed under the Mozilla Public Licence (MPL) v2.0 except for the files
# "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""EnPT EnMAP objects in sensor geometry."""

import logging
from types import SimpleNamespace
from typing import Tuple, Optional, List  # noqa: F401
from tempfile import TemporaryDirectory
from zipfile import ZipFile
import numpy as np
from os import path, makedirs
from glob import glob
import utm
from scipy.interpolate import interp2d
from geoarray import GeoArray

from ...utils.logging import EnPT_Logger
from .image_baseclasses import _EnMAP_Image
from ...model.metadata import EnMAP_Metadata_L1B_SensorGeo, EnMAP_Metadata_L1B_Detector_SensorGeo
from ...model.metadata import EnMAP_Metadata_L2A_MapGeo  # noqa: F401  # only used for type hint
from ...options.config import EnPTConfig
from ...processors.dead_pixel_correction import Dead_Pixel_Corrector
from ...processors.dem_preprocessing import DEM_Processor
from ...processors.spatial_transform import compute_mapCoords_within_sensorGeoDims

__author__ = ['Daniel Scheffler', 'Stéphane Guillaso', 'André Hollstein']


class EnMAP_Detector_SensorGeo(_EnMAP_Image):
    """Class representing a single detector of an EnMAP image (in sensor geometry).

    NOTE:
        - Inherits all attributes from _EnMAP_Image class.
        - All functionality that VNIR and SWIR detectors (sensor geometry) have in common is to be implemented here.

    Attributes:
        - to be listed here. Check help(_EnMAP_Detector_SensorGeo) in the meanwhile!

    """

    def __init__(self, detector_name: str, root_dir: str, config: EnPTConfig, logger=None, meta=None) -> None:
        """Get an instance of _EnMAP_Detector_SensorGeo.

        :param detector_name:   'VNIR' or 'SWIR'
        :param root_dir:
        :param logger:
        :param meta: import meta if already loaded
        """
        if detector_name not in ['VNIR', 'SWIR']:
            raise ValueError("'detector_name' must be 'VNIR' or 'SWIR'. Received %s!" % detector_name)

        self.cfg = config
        self._root_dir = root_dir
        self.detector_name = detector_name
        self.logger = logger or logging.getLogger()

        # get all attributes of base class "_EnMAP_Image"
        super(EnMAP_Detector_SensorGeo, self).__init__()

        # instance an empty metadata object
        if meta is None:
            self.detector_meta: EnMAP_Metadata_L1B_Detector_SensorGeo = \
                EnMAP_Metadata_L1B_Detector_SensorGeo(self.detector_name, config=self.cfg, logger=self.logger)
        else:
            self.detector_meta: EnMAP_Metadata_L1B_Detector_SensorGeo = meta

    def get_paths(self) -> SimpleNamespace:
        """
        Get all file paths associated with the current instance of EnMAP_Detector_SensorGeo
        These information are read from the detector_meta.
        :return: paths
        """
        self.paths.root_dir = self._root_dir
        self.paths.data = path.join(self._root_dir, self.detector_meta.filename_data)

        def path_or_None(filename):
            return path.join(self._root_dir, filename) if filename else None

        self.paths.mask_landwater = path_or_None(self.detector_meta.filename_mask_landwater)
        self.paths.mask_clouds = path_or_None(self.detector_meta.filename_mask_clouds)
        self.paths.mask_cloudshadow = path_or_None(self.detector_meta.filename_mask_cloudshadow)
        self.paths.mask_haze = path_or_None(self.detector_meta.filename_mask_haze)
        self.paths.mask_snow = path_or_None(self.detector_meta.filename_mask_snow)
        self.paths.mask_cirrus = path_or_None(self.detector_meta.filename_mask_cirrus)
        self.paths.deadpixelmap = path_or_None(self.detector_meta.filename_deadpixelmap)
        self.paths.quicklook = path_or_None(self.detector_meta.filename_quicklook)

        return self.paths

    def correct_dead_pixels(self):
        """Correct dead pixels with respect to the dead pixel mask."""
        algo = self.cfg.deadpix_P_algorithm
        method_spectral, method_spatial = self.cfg.deadpix_P_interp_spectral, self.cfg.deadpix_P_interp_spatial
        self.logger.info("Correcting dead pixels of %s detector...\n"
                         "Used algorithm: %s interpolation in the %s domain"
                         % (self.detector_name, algo, method_spectral if algo == 'spectral' else method_spatial))

        self.data = \
            Dead_Pixel_Corrector(algorithm=algo,
                                 interp_spectral=method_spectral,
                                 interp_spatial=method_spatial,
                                 logger=self.logger)\
            .correct(self.data, self.deadpixelmap)

    def get_preprocessed_dem(self):
        if self.cfg.path_dem:
            self.logger.info('Pre-processing DEM for %s...' % self.detector_name)
            DP = DEM_Processor(self.cfg.path_dem, enmapIm_cornerCoords=tuple(zip(self.detector_meta.lon_UL_UR_LL_LR,
                                                                                 self.detector_meta.lat_UL_UR_LL_LR)),
                               CPUs=self.cfg.CPUs)
            DP.fill_gaps()  # FIXME this will also be needed at other places

            R, C = self.data.shape[:2]
            if DP.dem.is_map_geo:
                lons = self.detector_meta.lons
                lats = self.detector_meta.lats

                if not (lons.ndim == 2 and lats.ndim == 2) and not (lons.ndim == 3 and lats.ndim == 3):
                    raise ValueError((lons.ndim, lats.ndim), 'Geolayer must be either 2- or 3-dimensional.')

                msg_bandinfo = ''
                if lons.ndim == 3:
                    # 3D geolayer (the usual case for EnMAP data provided by DLR)
                    lons = lons[:, :, 0]
                    lats = lats[:, :, 0]
                    msg_bandinfo = ' (using first band of %s geolayer)' % self.detector_name
                else:
                    # 2D geolayer (GFZ-internal test case)
                    # FIXME replace linear interpolation by native geolayers
                    if lons.shape != self.data.shape:
                        lons = self.detector_meta.interpolate_corners(*self.detector_meta.lon_UL_UR_LL_LR, nx=C, ny=R)
                    if lats.shape != self.data.shape:
                        lats = self.detector_meta.interpolate_corners(*self.detector_meta.lat_UL_UR_LL_LR, nx=C, ny=R)

                self.logger.info(('Transforming DEM to %s sensor geometry%s...' % (self.detector_name, msg_bandinfo)))
                self.dem = DP.to_sensor_geometry(lons=lons, lats=lats)
            else:
                self.dem = DP.dem

        return self.dem

    def append_new_image(self, img2: 'EnMAP_Detector_SensorGeo', n_lines: int = None) -> None:
        # TODO convert method to function?
        """Check if a second image matches with the first image and if so, append the given number of lines below.

        In this version we assume that the image will be added below. If it is the case, the method will create
        temporary files that will be used in the following.

        :param img2:
        :param n_lines: number of lines to be added from the new image
        :return: None
        """
        if not self.cfg.is_dummy_dataformat:
            basename_img1 = self.detector_meta.filename_data.split('-SPECTRAL_IMAGE')[0] + '::%s' % self.detector_name
            basename_img2 = img2.detector_meta.filename_data.split('-SPECTRAL_IMAGE')[0] + '::%s' % img2.detector_name
        else:
            basename_img1 = path.basename(self._root_dir)
            basename_img2 = path.basename(img2._root_dir)

        self.logger.info("Check new image for %s: %s " % (self.detector_name, basename_img2))

        distance_min = 27.0
        distance_max = 34.0

        # compute distance between image1 LL and image2 UL
        x1, y1, _, _ = utm.from_latlon(self.detector_meta.lat_UL_UR_LL_LR[2], self.detector_meta.lon_UL_UR_LL_LR[2])
        x2, y2, _, _ = utm.from_latlon(img2.detector_meta.lat_UL_UR_LL_LR[0], img2.detector_meta.lon_UL_UR_LL_LR[0])
        distance_left = np.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)

        # compute distance between image1 LR and image2 UR
        x1, y1, _, _ = utm.from_latlon(self.detector_meta.lat_UL_UR_LL_LR[3], self.detector_meta.lon_UL_UR_LL_LR[3])
        x2, y2, _, _ = utm.from_latlon(img2.detector_meta.lat_UL_UR_LL_LR[1], img2.detector_meta.lon_UL_UR_LL_LR[1])
        distance_right = np.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)

        if distance_min < distance_left < distance_max and distance_min < distance_right < distance_max:
            self.logger.info("Append new image to %s: %s" % (self.detector_name, basename_img2))
        else:
            self.logger.warning("%s and %s don't fit to be appended." % (basename_img1, basename_img2))
            return

        # set new number of line
        n_lines = n_lines or img2.detector_meta.nrows

        if n_lines > img2.detector_meta.nrows:
            self.logger.warning("n_lines (%s) exceeds the total number of line of second image" % n_lines)
            self.logger.warning("Set to the image number of line")
            n_lines = img2.detector_meta.nrows

        if n_lines < 50:  # TODO: determine these values
            self.logger.warning("A minimum of 50 lines is required, only %s were selected" % n_lines)
            self.logger.warning("Set the number of line to 50")
            n_lines = 50

        self.detector_meta.nrows += n_lines

        # Compute new lower coordinates
        if not self.cfg.is_dummy_dataformat:
            img2_cornerCoords = tuple(zip(img2.detector_meta.lon_UL_UR_LL_LR,
                                          img2.detector_meta.lat_UL_UR_LL_LR))
            dem_validated = DEM_Processor(img2.cfg.path_dem,
                                          enmapIm_cornerCoords=img2_cornerCoords).dem
            LL, LR = compute_mapCoords_within_sensorGeoDims(
                sensorgeoCoords_YX=[(n_lines - 1, 0),  # LL
                                    (n_lines - 1, img2.detector_meta.ncols - 1)],  # LR
                rpc_coeffs=list(img2.detector_meta.rpc_coeffs.values())[0],  # RPC coeffs of first band of the detector
                dem=dem_validated,
                enmapIm_cornerCoords=img2_cornerCoords,
                enmapIm_dims_sensorgeo=(img2.detector_meta.nrows, img2.detector_meta.ncols)
            )

            self.detector_meta.lon_UL_UR_LL_LR[2], self.detector_meta.lat_UL_UR_LL_LR[2] = LL
            self.detector_meta.lon_UL_UR_LL_LR[3], self.detector_meta.lat_UL_UR_LL_LR[3] = LR
        else:
            # lats
            interp_lats = interp2d(x=[0, 1],
                                   y=[0, 1],
                                   z=[[img2.detector_meta.lat_UL_UR_LL_LR[0], img2.detector_meta.lat_UL_UR_LL_LR[1]],
                                      [img2.detector_meta.lat_UL_UR_LL_LR[2], img2.detector_meta.lat_UL_UR_LL_LR[3]]],
                                   kind='linear')
            self.detector_meta.lat_UL_UR_LL_LR[2] = np.array(interp_lats(0, int(n_lines / img2.detector_meta.nrows)))[0]
            self.detector_meta.lat_UL_UR_LL_LR[3] = np.array(interp_lats(1, int(n_lines / img2.detector_meta.nrows)))[0]
            self.detector_meta.lats = self.detector_meta.interpolate_corners(*self.detector_meta.lat_UL_UR_LL_LR,
                                                                             self.detector_meta.ncols,
                                                                             self.detector_meta.nrows)
            # lons
            interp_lons = interp2d(x=[0, 1],
                                   y=[0, 1],
                                   z=[[img2.detector_meta.lon_UL_UR_LL_LR[0], img2.detector_meta.lon_UL_UR_LL_LR[1]],
                                      [img2.detector_meta.lon_UL_UR_LL_LR[2], img2.detector_meta.lon_UL_UR_LL_LR[3]]],
                                   kind='linear')
            self.detector_meta.lon_UL_UR_LL_LR[2] = np.array(interp_lons(0, int(n_lines / img2.detector_meta.nrows)))[0]
            self.detector_meta.lon_UL_UR_LL_LR[3] = np.array(interp_lons(1, int(n_lines / img2.detector_meta.nrows)))[0]
            self.detector_meta.lons = self.detector_meta.interpolate_corners(*self.detector_meta.lon_UL_UR_LL_LR,
                                                                             self.detector_meta.ncols,
                                                                             self.detector_meta.nrows)

        # append the raster data
        self.data = np.append(self.data, img2.data[0:n_lines, :, :], axis=0)

        # only append masks for the VNIR as they are only provided in VNIR sensor geometry
        if self.detector_name == 'VNIR':
            for attrName in ['mask_landwater', 'mask_clouds', 'mask_cloudshadow',
                             'mask_haze', 'mask_snow', 'mask_cirrus']:
                arr_img1 = getattr(self, attrName)
                arr_img2 = getattr(img2, attrName)

                if arr_img1 is not None and arr_img2 is not None:
                    setattr(self, attrName, np.append(arr_img1, arr_img2[0:n_lines, :], axis=0))
                else:
                    # this mainly applies to the dummy data format that does not have all mask files
                    self.logger.warning("Could not append the '%s' attribute "
                                        "as it does not exist in the current image." % attrName)

        if not self.cfg.is_dummy_dataformat:
            self.deadpixelmap = np.append(self.deadpixelmap, img2.deadpixelmap[0:n_lines, :], axis=0)

        # NOTE: We leave the quicklook out here because merging the quicklook of adjacent scenes might cause a
        #       brightness jump that can be avoided by recomputing the quicklook after DN/radiance conversion.

    def DN2TOARadiance(self):
        """Convert DNs to TOA radiance.

        Convert the radiometric unit of _EnMAP_Detector_SensorGeo.data from digital numbers to top-of-atmosphere
        radiance.

        :return: None
        """
        # TODO move to processors.radiometric_transform?
        if self.detector_meta.unitcode == 'DN':
            self.logger.info('Converting DN values to radiance [mW/m^2/sr/nm] for %s detector...' % self.detector_name)

            if self.detector_meta.l_min is not None and self.detector_meta.l_max is not None:
                # Lλ = (LMINλ + ((LMAXλ - LMINλ)/(QCALMAX-QCALMIN)) * (QCAL-QCALMIN))
                # FIXME this asserts LMIN and LMAX in mW m^-2 sr^-1 nm^-1

                QCALMIN = 1
                QCALMAX = 2 ** 16  # 65535 (16-bit maximum value)
                LMIN = self.detector_meta.l_min
                LMAX = self.detector_meta.l_max
                QCAL = self.data[:]

                self.data = ((LMAX - LMIN)/(QCALMAX - QCALMIN)) * (QCAL - QCALMIN) + LMIN

            elif self.detector_meta.gains is not None and self.detector_meta.offsets is not None:
                # Lλ = QCAL * GAIN + OFFSET
                # NOTE: - DLR provides gains between 2000 and 10000, so we have to DEVIDE by gains
                #       - DLR gains / offsets are provided in W/m2/sr/nm, so we have to multiply by 1000 to get
                #         mW/m2/sr/nm as needed later
                self.data = 1000 * (self.data[:] * self.detector_meta.gains + self.detector_meta.offsets)

            else:
                raise ValueError("Neighter 'l_min'/'l_max' nor 'gains'/'offsets' "
                                 "are available for radiance computation.")

            self.detector_meta.unit = "mW m^-2 sr^-1 nm^-1"
            self.detector_meta.unitcode = "TOARad"
        else:
            self.logger.warning(
                "No DN to Radiance conversion is performed because unitcode is not DN (found: {code}).".format(
                    code=self.detector_meta.unitcode)
            )

    def save_raster_attributes(self, attrNames: List[str], outputDir: str):
        """Save the specified raster attributes to the given output directory.

        :param attrNames:   list of attribute names to be saved
        :param outputDir:   output directory
        """
        for attrName in attrNames:
            attr = getattr(self, attrName)  # type: GeoArray
            fN = getattr(self.detector_meta, 'filename_%s' % attrName)

            if attr is not None:
                attr.save(path.join(outputDir, fN), fmt="GTiff")
            else:
                self.logger.warning("Could not save the %s attribute '%s' as it does not exist in the current image."
                                    % (self.detector_name, attrName))


class EnMAP_VNIR_SensorGeo(EnMAP_Detector_SensorGeo):
    def __init__(self, root_dir: str, config: EnPTConfig, logger=None, meta=None) -> None:
        super().__init__(detector_name='VNIR', root_dir=root_dir, config=config, logger=logger, meta=meta)

    def read_masks(self):
        """
        Read the L1B masks.
        """
        self.logger.info('Reading image masks in VNIR sensor geometry.')

        # water mask (0=backgr.; 1=land; 2=water)
        if self.paths.mask_landwater:
            self.mask_landwater = GeoArray(self.paths.mask_landwater)[:]

        # cloud mask (0=none; 1=cloud)
        if self.paths.mask_clouds:
            self.mask_clouds = GeoArray(self.paths.mask_clouds)[:] == 1

        # cloud shadow mask (0=none; 1=cloud shadow)
        if self.paths.mask_cloudshadow:
            self.mask_cloudshadow = GeoArray(self.paths.mask_cloudshadow)[:] == 1

        # haze mask (0=none; 1=haze)
        if self.paths.mask_landwater:
            self.mask_haze = GeoArray(self.paths.mask_haze)[:] == 1

        # snow mask (0=none; 1=snow)
        if self.paths.mask_snow:
            self.mask_snow = GeoArray(self.paths.mask_snow)[:] == 1

        # cirrus mask (0=none; 1=thin, 2=medium, 3=thick)
        if self.paths.mask_cirrus:
            self.mask_cirrus = GeoArray(self.paths.mask_cirrus)[:]


class EnMAP_SWIR_SensorGeo(EnMAP_Detector_SensorGeo):
    def __init__(self, root_dir: str, config: EnPTConfig, logger=None, meta=None) -> None:
        super().__init__(detector_name='SWIR', root_dir=root_dir, config=config, logger=logger, meta=meta)

    def __getattribute__(self, item):  # called whenever an instance attribute is accessed
        if item in ['mask_landwater', 'mask_clouds', 'mask_cloudshadow',
                    'mask_haze', 'mask_snow', 'mask_cirrus'] \
                and getattr(self, '_%s' % item) is None:
            self.logger.warning('The %s is not yet available in SWIR sensor geometry. '
                                'Use EnMAP_SWIR_SensorGeo.transform_vnir_to_swir_raster() to set it with a '
                                'transformed version of the one provided in VNIR sensor geometry.' % item)
        return super().__getattribute__(item)

    def transform_vnir_to_swir_raster(self,
                                      array_vnirsensorgeo: np.ndarray,
                                      vnir_lons: np.ndarray,
                                      vnir_lats: np.ndarray,
                                      vnir_epsg: int) -> np.ndarray:
        """Set a SWIR raster attribute with a VNIR attribute transformed to SWIR sensor geometry.

        :param array_vnirsensorgeo: source array in VNIR sensor geometry wo be transformed
        :param vnir_lons:           longitude geolayer array of the VNIR
        :param vnir_lats:           latitude geolayer array of the VNIR
        :param vnir_epsg:           EPSG code of the VNIR when transformed to map geometry
        """
        if self.detector_meta.lons is None or self.detector_meta.lats is None:
            raise RuntimeError('The SWIR geolayer must be computed first '
                               'to transform arrays from VNIR to SWIR sensor geometry.')

        from ...processors.spatial_transform import VNIR_SWIR_SensorGeometryTransformer
        VS_SGT = VNIR_SWIR_SensorGeometryTransformer(lons_vnir=vnir_lons[:, :, 0],
                                                     lats_vnir=vnir_lats[:, :, 0],
                                                     lons_swir=self.detector_meta.lons[:, :, 0],
                                                     lats_swir=self.detector_meta.lats[:, :, 0],
                                                     prj_vnir=vnir_epsg,
                                                     prj_swir=self.detector_meta.epsg_ortho,
                                                     res_vnir=(30, 30),
                                                     res_swir=(30, 30),
                                                     resamp_alg='nearest',
                                                     # radius_of_influence=45,
                                                     nprocs=self.cfg.CPUs
                                                     )
        return VS_SGT.transform_sensorgeo_VNIR_to_SWIR(array_vnirsensorgeo)


class EnMAPL1Product_SensorGeo(object):
    """Class for EnPT EnMAP object in sensor geometry.

    Attributes:
        - logger:
            - logging.Logger instance or subclass instance
        - vnir
            - instance of EnMAP_VNIR_SensorGeo class
        - swir
            - instance of EnMAP_SWIR_SensorGeo class
        - paths:
            - paths belonging to the EnMAP product
        - meta:
            - instance of EnMAP_Metadata_SensorGeo class
        - detector_attrNames:
            - list of attribute names for VNIR and SWIR detectors,

    """

    def __init__(self, root_dir: str, config: EnPTConfig, logger=None):
        """Get instance of EnPT EnMAP object in sensor geometry.

        :param root_dir:    Root directory of EnMAP Level-1B product
        :param config:      EnPT configuration object
        :param logger:      None or logging instance to be appended to EnMAPL1Product_SensorGeo instance
                            (If None, a default EnPT_Logger is used).
        """
        # protected attributes
        self._logger = None

        # populate attributes
        self.cfg = config
        if logger:
            self.logger = logger

        # Read metadata here in order to get all information needed by to get paths.
        if not self.cfg.is_dummy_dataformat:
            self.meta = EnMAP_Metadata_L1B_SensorGeo(glob(path.join(root_dir, "*METADATA.XML"))[0],
                                                     config=self.cfg, logger=self.logger)
        else:
            self.meta = EnMAP_Metadata_L1B_SensorGeo(glob(path.join(root_dir, "*_header.xml"))[0],
                                                     config=self.cfg, logger=self.logger)
        self.meta.read_metadata()

        # define VNIR and SWIR detector
        self.detector_attrNames = ['vnir', 'swir']
        self.vnir = EnMAP_VNIR_SensorGeo(root_dir, config=self.cfg, logger=self.logger, meta=self.meta.vnir)
        self.swir = EnMAP_SWIR_SensorGeo(root_dir, config=self.cfg, logger=self.logger, meta=self.meta.swir)

        # Get the paths according information delivered in the metadata
        self.paths = self.get_paths()

        # associate raster attributes with file links (raster data is read lazily / on demand)
        self.vnir.data = self.paths.vnir.data
        self.vnir.read_masks()
        self.swir.data = self.paths.swir.data

        try:
            self.vnir.deadpixelmap = self.paths.vnir.deadpixelmap
            self.swir.deadpixelmap = self.paths.swir.deadpixelmap
        except ValueError:
            self.logger.warning("Unexpected dimensions of dead pixel mask. Setting all pixels to 'normal'.")
            self.vnir.deadpixelmap = np.zeros(self.vnir.data.shape)
            self.swir.deadpixelmap = np.zeros(self.swir.data.shape)

        # NOTE: We leave the quicklook out here because merging the quicklook of adjacent scenes might cause a
        #       brightness jump that can be avoided by recomputing the quicklook after DN/radiance conversion.

        # compute radiance
        self.DN2TOARadiance()

    def get_paths(self):
        """
        Get all file paths associated with the current instance of EnMAPL1Product_SensorGeo

        :return: paths.SimpleNamespace()
        """
        paths = SimpleNamespace()
        paths.vnir = self.vnir.get_paths()
        paths.swir = self.swir.get_paths()
        paths.root_dir = self.meta.rootdir
        paths.metaxml = self.meta.path_xml

        return paths

    @property
    def logger(self) -> EnPT_Logger:
        """Get a an instance of enpt.utils.logging.EnPT_Logger.

        NOTE:
            - The logging level will be set according to the user inputs of EnPT.
            - The path of the log file is directly derived from the attributes of the _EnMAP_Image instance.

        Usage:
            - get the logger:
                logger = self.logger
            - set the logger
                self.logger = logging.getLogger()  # NOTE: only instances of logging.Logger are allowed here
            - delete the logger:
                del self.logger  # or "self.logger = None"

        :return: EnPT_Logger instance
        """
        if self._logger and self._logger.handlers[:]:
            return self._logger
        else:
            basename = path.splitext(path.basename(self.cfg.path_l1b_enmap_image))[0]
            path_logfile = path.join(self.cfg.output_dir, basename + '.log') \
                if self.cfg.create_logfile and self.cfg.output_dir else ''
            self._logger = EnPT_Logger('log__' + basename, fmt_suffix=None, path_logfile=path_logfile,
                                       log_level=self.cfg.log_level, append=False)

            return self._logger

    @logger.setter
    def logger(self, logger: logging.Logger):
        assert isinstance(logger, logging.Logger) or logger in ['not set', None], \
            "%s.logger can not be set to %s." % (self.__class__.__name__, logger)

        # save prior logs
        # if logger is None and self._logger is not None:
        #     self.log += self.logger.captured_stream
        self._logger = logger

    @property
    def log(self) -> str:
        """Return a string of all logged messages until now.

        NOTE: self.log can also be set to a string.
        """
        return self.logger.captured_stream

    @log.setter
    def log(self, string: str):
        assert isinstance(string, str), "'log' can only be set to a string. Got %s." % type(string)
        self.logger.captured_stream = string

    # @classmethod
    # def from_L1B_provider_data(cls, path_enmap_image: str, config: EnPTConfig=None) -> EnMAPL1Product_SensorGeo:
    #     """
    #
    #     :param path_enmap_image:
    #     :param config:
    #     :return:
    #     """
    #     # input validation
    #     if not path.isdir(path_enmap_image) and \
    #        not (path.exists(path_enmap_image) and path_enmap_image.endswith('.zip')):
    #         raise ValueError("The parameter 'path_enmap_image' must be a directory or the path to an existing zip "
    #                          "archive.")
    #
    #     # extract L1B image archive if needed
    #     if path_enmap_image.endswith('.zip'):
    #         path_enmap_image = self.extract_zip_archive(path_enmap_image)
    #         if not path.isdir(path_enmap_image):
    #             raise NotADirectoryError(path_enmap_image)
    #
    #     # run the reader
    #     from ..io.reader import L1B_Reader
    #     RD = L1B_Reader(config=config)
    #     L1_obj = RD.read_inputdata(path_enmap_image, observation_time=datetime(2015, 12, 7, 10))
    #
    #     return L1_obj

    def DN2TOARadiance(self):
        """Convert self.vnir.data and self.swir.data from digital numbers to top-of-atmosphere radiance.

        :return: None
        """
        if self.vnir.detector_meta.unitcode != 'TOARad':
            self.vnir.DN2TOARadiance()
            self.meta.vnir.unitcode = self.vnir.detector_meta.unitcode  # FIXME possible duplicate?
            self.meta.vnir.unit = self.vnir.detector_meta.unit  # FIXME possible duplicate?
        if self.swir.detector_meta.unitcode != 'TOARad':
            self.swir.DN2TOARadiance()
            self.meta.swir.unitcode = self.swir.detector_meta.unitcode
            self.meta.swir.unit = self.swir.detector_meta.unit

    def append_new_image(self, root_dir, n_line_ext):
        """Append a second EnMAP Level-1B product below the last line of the current L1B product.

        NOTE:   We create new files that will be saved into a temporary directory and their path will be stored in the
                paths of self. We assume that the dead pixel map does not change between two adjacent images.

        :param root_dir:    root directory of EnMAP Level-1B product to be appended
        :param n_line_ext:  number of lines to be added from the new image
        :return:
        """
        l1b_ext_obj = EnMAPL1Product_SensorGeo(root_dir, config=self.cfg)

        self.vnir.append_new_image(l1b_ext_obj.vnir, n_line_ext)
        self.swir.append_new_image(l1b_ext_obj.swir, n_line_ext)

    def calc_snr_from_radiance(self):
        """Compute EnMAP SNR from radiance data.

        SNR equation:    SNR = p0 + p1 * LTOA + p2 * LTOA ^ 2   [W/(m^2 sr nm)].
        """
        with TemporaryDirectory(dir=self.cfg.working_dir) as tmpDir, \
                ZipFile(self.cfg.path_l1b_snr_model, "r") as zf:

            zf.extractall(tmpDir)

            if self.meta.vnir.unitcode == 'TOARad':
                self.vnir.detector_meta.calc_snr_from_radiance(rad_data=self.vnir.data, dir_snr_models=tmpDir)

            if self.meta.swir.unitcode == 'TOARad':
                self.swir.detector_meta.calc_snr_from_radiance(rad_data=self.swir.data, dir_snr_models=tmpDir)

    def correct_dead_pixels(self):
        """Correct dead pixels of both detectors."""
        self.vnir.correct_dead_pixels()
        self.swir.correct_dead_pixels()

    # def correct_VNIR_SWIR_shift(self):
    #     # use first geolayer bands for VNIR and SWIR
    #     Vlons, Vlats = self.vnir.detector_meta.lons[:, :, 0], self.vnir.detector_meta.lats[:, :, 0]
    #     Slons, Slats = self.swir.detector_meta.lons[:, :, 0], self.swir.detector_meta.lats[:, :, 0]
    #
    #     # get corner coordinates of VNIR and SWIR according to geolayer
    #     def get_coords(lons, lats):
    #         return tuple([(lons[Y, X], lats[Y, X]) for Y, X in [(0, 0), (0, -1), (-1, 0), (-1, -1)]])
    #
    #     VUL, VUR, VLL, VLR = get_coords(Vlons, Vlats)
    #     SUL, SUR, SLL, SLR = get_coords(Slons, Slats)
    #
    #     # get map coordinates of VNIR/SWIR overlap
    #     ovUL = max(VUL[0], SUL[0]), min(VUL[1], SUL[1])
    #     ovUR = min(VUR[0], SUR[0]), min(VUR[1], SUR[1])
    #     ovLL = max(VLL[0], SLL[0]), max(VLL[1], SLL[1])
    #     ovLR = min(VLR[0], SLR[0]), max(VLR[1], SLR[1])
    #
    #     # find nearest image positions for VNIR and SWIR to the map coordinates of the VNIR/SWIR overlap
    #     def nearest_imCoord(lons_arr, lats_arr, lon, lat):
    #         dists = np.sqrt((lons_arr - lon) ** 2 + (lats_arr - lat) ** 2)
    #         row, col = np.unravel_index(dists.argmin(), dists.shape)
    #
    #         return row, col
    #
    #     overlapImVNIR = tuple([nearest_imCoord(Vlons, Vlats, *ovCoords) for ovCoords in [ovUL, ovUR, ovLL, ovLR]])
    #     overlapImSWIR = tuple([nearest_imCoord(Slons, Slats, *ovCoords) for ovCoords in [ovUL, ovUR, ovLL, ovLR]])
    #
    #     raise NotImplementedError()  # FIXME
    #     self.vnir.data.get_mapPos()  # FIXME

    def get_preprocessed_dem(self):
        self.vnir.get_preprocessed_dem()
        self.swir.get_preprocessed_dem()

    def transform_vnir_to_swir_raster(self, attrName):
        """Set the specified SWIR raster attribute with a VNIR attribute transformed to SWIR sensor geometry.

        :param attrName:    name of the attribute to be set
        """
        self.logger.info("Transforming the '%s' attribute from VNIR to SWIR sensor geometry." % attrName)

        vnir_rasterAttr = getattr(self.vnir, attrName)

        if self.meta.vnir.lons is None or self.meta.vnir.lats is None or \
           self.meta.swir.lons is None or self.meta.swir.lats is None:
            raise RuntimeError('The VNIR/SWIR geolayers must be computed first '
                               'to transform arrays from VNIR to SWIR sensor geometry.')

        if vnir_rasterAttr is None:
            raise RuntimeError("%s.vnir.%s has not yet been set." % (self.__class__.__name__, attrName))

        attr_transformed = self.swir.transform_vnir_to_swir_raster(array_vnirsensorgeo=np.array(vnir_rasterAttr),
                                                                   vnir_lons=self.meta.vnir.lons,
                                                                   vnir_lats=self.meta.vnir.lats,
                                                                   vnir_epsg=self.meta.vnir.epsg_ortho)
        setattr(self.swir, attrName, attr_transformed)

    def run_AC(self):
        from ...processors.atmospheric_correction import AtmosphericCorrector
        AC = AtmosphericCorrector(config=self.cfg)
        AC.run_ac(self)

    def save(self, outdir: str, suffix="") -> str:
        """Save the product to disk using almost the same input format.

        NOTE: Radiance is saved instead of DNs to avoid a radiometric jump between concatenated images.

        :param outdir: path to the output directory
        :param suffix: suffix to be appended to the output filename (???)
        :return: root path (root directory) where products were written
        """
        product_dir = path.join(path.abspath(outdir),
                                "{name}{suffix}".format(name=self.meta.scene_basename, suffix=suffix))

        self.logger.info("Write product to: %s" % product_dir)
        makedirs(product_dir, exist_ok=True)

        # write the VNIR
        self.vnir.save_raster_attributes(['data', 'mask_landwater', 'mask_clouds', 'mask_cloudshadow',
                                          'mask_haze', 'mask_snow', 'mask_cirrus', 'deadpixelmap'],
                                         product_dir)

        # FIXME we could also write the quicklook included in DLR L1B format instead of generating an own one
        self.vnir.generate_quicklook(bands2use=self.vnir.detector_meta.preview_bands) \
            .save(path.join(product_dir, path.splitext(self.meta.vnir.filename_quicklook)[0] + '.png'), fmt='PNG')

        # write the SWIR
        # NOTE: The masks only exist in VNIR sensor geometry (would have to be transformed before to match for the SWIR)
        self.swir.save_raster_attributes(['data', 'deadpixelmap'], product_dir)

        self.swir.generate_quicklook(bands2use=self.swir.detector_meta.preview_bands) \
            .save(path.join(product_dir, path.splitext(self.meta.swir.filename_quicklook)[0] + '.png'), fmt='PNG')

        # Update the xml file
        metadata_string = self.meta.to_XML()
        with open(product_dir + path.sep + path.basename(self.paths.metaxml), "w") as xml_file:
            xml_file.write(metadata_string)

        self.logger.info("L1B product successfully written!")

        return product_dir
